/* ARP-mangling code to bring switches in flooding mode 
 * by Alex Stieglecker <alex_oz@gmx.net>
 * based on code of Harald Welte <laforge@netfilter.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 */
 
#ifndef __KERNEL__
#  define __KERNEL__
#endif
#ifndef MODULE
#  define MODULE
#endif

/*
 * Inclusions related to version problems. They *might* go after
 * <linux/module.h>, unless a symbol table is exported. All in all,
 * better putting them first anyways.
 */
#include <linux/config.h> /* retrieve the CONFIG_* macros */
#if defined(CONFIG_MODVERSIONS) && !defined(MODVERSIONS)
#  define MODVERSIONS /* force it on */
#endif

#ifdef MODVERSIONS
#  include <linux/modversions.h>
#endif
/* end version stuff */

 
#include <linux/module.h>
#include <linux/moduleparam.h>
#include <linux/config.h>
#include <linux/proc_fs.h>
#include <linux/skbuff.h>
#include <linux/if_arp.h>
#include <linux/if_packet.h>
#include <linux/inetdevice.h>
#include <linux/netdevice.h>
#include <linux/rtnetlink.h>
#include <linux/ip.h>

#include <linux/netfilter_ipv4/ip_conntrack.h>
#include <linux/proc_fs.h>


#include <net/checksum.h>

#include <linux/netfilter_arp.h>

#define ARPFAKE_VERSION "0.0.1"

#define DEBUG_ARPFAKE

#ifdef DEBUG_ARPFAKE
#define DEBUGP	printk
#else
#define DEBUGP
#endif

#define NF_IP_PRI_ARPFAKE -450 /* should be in netfilter_ipv4.h, NF_IP_PRI_SELECT is higher */

#define MAX_PDE_LEN 32         /* maximum length of procdir-entries */

#define MAX_PARAM_LEN 80

#define is_mc_ip(ip) (!(~((ip) | 0x1FFFFFFF)))

#ifdef CONFIG_PROC_FS
static int arpfake_proc_read( char *buf, char **start, off_t offset,
                              int size, int *eof, void *data);
static int arpfake_proc_write( struct file *instanz, const char __user *userbuffer,                               unsigned long count, void *data );
#endif

MODULE_LICENSE("GPL");
MODULE_AUTHOR("Alexander Stieglecker <alez_oz@gmx.net>");
MODULE_DESCRIPTION("ARP mangling code, to bring switches in flooding mode");

static int mangle = 0;
MODULE_PARM_DESC( mangle, " 0 = mangle no ARP-req/resp; 1 = mangle all ARP-req/resp" );
module_param(mangle, bool, 0);

/* linked list of 'fake-items' (information how to fake an arp request/reply) */
struct fake {
	struct fake		*next;		/* list of all configs */
	atomic_t		refcount;	/* reference count */
	u_int8_t		mmacaddr[6];	/* multicast mac */
	u_int32_t		addr;		/* address of fake, 
						 * identifies fake as a 
						 * unique element in list */
	u_int8_t		active;		/* activates faking */

#ifdef CONFIG_PROC_FS
	struct proc_dir_entry *pde;		/* proc dir entry */
#endif

};

/* head of fake */
static struct fake *fake_head = NULL;
/* protects access to fake */
DECLARE_RWLOCK(fake_lock);

#ifdef CONFIG_PROC_FS
static struct proc_dir_entry *arpfake_procdir;   /* procdir entry for all running interfaces */
#endif


/* hardcoded for 48bit ethernet and 32bit ipv4 addresses */
struct arp_payload {
	u_int8_t src_hw[ETH_ALEN];
	u_int32_t src_ip;
	u_int8_t dst_hw[ETH_ALEN];
	u_int32_t dst_ip;
} __attribute__ ((packed));

#ifdef DEBUG_ARPFAKE
static void arp_print(struct arp_payload *payload) 
{
#define HBUFFERLEN 30
	char hbuffer[HBUFFERLEN];
	int j,k;
	const char hexbuf[]= "0123456789abcdef";

	for (k=0, j=0; k < HBUFFERLEN-3 && j < ETH_ALEN; j++) {
		hbuffer[k++]=hexbuf[(payload->src_hw[j]>>4)&15];
		hbuffer[k++]=hexbuf[payload->src_hw[j]&15];
		hbuffer[k++]=':';
	}
	hbuffer[--k]='\0';

	printk("src %u.%u.%u.%u@%s, dst %u.%u.%u.%u\n", 
		NIPQUAD(payload->src_ip), hbuffer,
		NIPQUAD(payload->dst_ip));
}
#endif /* ARPFAKE_DEBUG */

#ifdef CONFIG_PROC_FS

/** Splits the string 'str' into tokens. The array of characterarrays contains
  * the pointers to several tokens within the string. After splitting 'str' is
  * destroyed.
  * 'tokenarr' is an array of characterarrays (space must be provided by caller)
  * 'sep' represents the seperator within the strings.
  * 'tokenarrlen' is the maximum number of tokens tokenarr can contain
  * 'tokennum' contains the actual number of tokens after execution 
  *
  * returns 1 if the splitting suceeded
  * returns 0 if the string was malformed.
  */
static int split_str(char *str, char sep, int *tokennum, char **tokenarr, int maxtokens)
{
	int i;
	*tokennum=1;
	
	if(!str[0] || str[0]==sep || maxtokens < 1) {
		*tokennum=0;
		return 0;
	}
	
	/* empty string is valid */
	if(str[0]=='\0') {
		*tokennum=0;
		return 1;
	}
	
	tokenarr[0] = str;
	for(i = 0; str[i] != '\0'; i++) {
		if(str[i] == sep) {
			if(str[i+1]=='\0' || str[i+1]== sep)
				return 0;
			str[i] = '\0';
			if(*tokennum >= maxtokens)
				return 0;
			tokenarr[(*tokennum)++] = str+i+1;
		}
	}
	
	return 1;
}

/** Converts the string representation of an mac into an array of uint8 numbers
  * (host order)
  * After the conversion mac is destroyed.
  * returns 1 if the conversion suceeded
  * returns -EINVAL if the string was malformed.
  */
int parse_mac(char *str_mac, u_int8_t *mac)
{
	int tokennum, i, tmp;
	char *tokenarr[6], *invchar;

	if( !str_mac ) {
		return -EINVAL;
	}
	
	if(!split_str(str_mac, ':', &tokennum, tokenarr, 6)) {
		return -EINVAL;
	}
	
	if(tokennum != 6) {
		return -EINVAL;
	}

	for(i=0; i<6; i++) {
		tmp = simple_strtoul(tokenarr[i], &invchar, 16);
		if(invchar[0] != '\0' || tmp > 0xff) {
			return -EINVAL;
		}
		mac[i] = tmp;
	}

	return 1;
}
#endif /* CONFIG_PROC_FS */

/*********************************************************************** 
 * LINKED LIST FUNCTIONS 
 ***********************************************************************/
 
static int
fake_print(struct fake *elem, char *buf, int size)
{
	int bytes_written;
	
	READ_LOCK(&fake_lock);
	
	if( !elem ) {
		READ_UNLOCK(&fake_lock);
		return -EINVAL;
	}
	
	bytes_written = snprintf( buf, size, "%02X:%02X:%02X:%02X:%02X:%02X|%s\n", 
	                           elem->mmacaddr[0], elem->mmacaddr[1], elem->mmacaddr[2],
	                           elem->mmacaddr[3], elem->mmacaddr[4], elem->mmacaddr[5],
	                           elem->active ? "enabled" : "disabled" );
	
	READ_UNLOCK(&fake_lock);

	return size >= bytes_written ? bytes_written : -ENOMEM;
}

/** adds the multicast mac address 'mmacaddr' to the first interface, owning
  * the ip-addr 'addr'. 
  */
static void
enable_mc(u_int32_t addr, u_int8_t *mmacaddr)
{
	struct net_device *dev = ip_dev_find(ntohl(addr));
	if(dev) dev_mc_add(dev, mmacaddr, ETH_ALEN, 0);
	else printk(KERN_WARNING"arp_arpfake: couldn't add multicast MAC\n");
}

/** adds the multicast mac address 'mmacaddr' to the first interface, owning
  * the ip-addr 'addr'. 
  */
static void
disable_mc(u_int32_t addr, u_int8_t *mmacaddr)
{
	struct net_device *dev = ip_dev_find(ntohl(addr));
	if(dev) dev_mc_delete(dev, mmacaddr, ETH_ALEN, 0);
	else printk(KERN_WARNING"arp_arpfake: couldn't remove multicast MAC\n");
}

/** adds a fake-element
 *  returns 0 if there where no errors
 *  returns -ENOMEM of there is no memory left
 */
static unsigned int
fake_add(u_int8_t *mmacaddr, u_int32_t addr, u_int8_t active)
{
	struct fake *actelem;
	struct fake *newelem;

#ifdef CONFIG_PROC_FS
	char buffer[MAX_PDE_LEN];
#endif

	newelem = kmalloc(sizeof(*newelem), GFP_ATOMIC);
	if(!newelem) {
		return -ENOMEM;
	}
	
	newelem->next = NULL;
	atomic_set(&newelem->refcount, 1);
	memcpy(newelem->mmacaddr, mmacaddr, 6 * sizeof(u_int8_t));
	newelem->addr = addr;
	newelem->active = active;

#ifdef CONFIG_PROC_FS
	/* create proc dir entry */
	sprintf(buffer, "%u.%u.%u.%u", HIPQUAD(addr));
	/* FIXME: exchange proc_net with arpfake_procdir */
	newelem->pde = create_proc_entry(buffer, S_IWUSR|S_IRUSR, proc_net);
	if (!newelem->pde) {
		kfree(newelem);
		return -ENOMEM;
	}
	
	newelem->pde->data = newelem;
	newelem->pde->read_proc  = arpfake_proc_read;
	newelem->pde->write_proc = arpfake_proc_write;

#endif

	WRITE_LOCK(&fake_lock);
	if(!fake_head) {
		fake_head = newelem;
		
		enable_mc(newelem->addr, newelem->mmacaddr);
		
		WRITE_UNLOCK(&fake_lock);
		return 1;
	}
	
	actelem = fake_head;
	
	while(actelem->next != NULL) {
		actelem = actelem->next;
	}
	
	actelem->next = newelem;
	
	enable_mc(newelem->addr, newelem->mmacaddr);
	
	WRITE_UNLOCK(&fake_lock);
	return 1;
}

/** only called from fake_put */
static unsigned int 
fake_del(struct fake *fake_elem)
{
	int ret = 1;
	struct fake *actelem;

	MUST_BE_WRITE_LOCKED(&fake_lock);

	if( !fake_elem ) {
		return -EINVAL;
	}

	if( !fake_head ) {
		ret = -EINVAL;
		goto free_elem;
	}

	
	actelem = fake_head;
	
	if(fake_head == fake_elem) {
		fake_head = fake_head->next;
		goto free_elem;
	}

	while(actelem->next != fake_elem) {
		if(actelem->next == NULL) {
			ret = -EINVAL;
			goto free_elem;
		}
		actelem = actelem->next;
	}

	actelem->next = fake_elem->next;

free_elem:
	disable_mc(fake_elem->addr, fake_elem->mmacaddr);

#ifdef CONFIG_PROC_FS
		if( !fake_elem->pde ) {
			printk(KERN_ERR "arp_arpfake: could not remove proc entry of fake\n");
		}
		else 
			remove_proc_entry(fake_elem->pde->name,
		                          fake_elem->pde->parent);
#endif  /* CONFIG_PROC_FS */
	
	kfree(fake_elem);
	return ret;

}

/** increments refcount of element f */
static inline void
fake_get(struct fake *f) {
	atomic_inc(&f->refcount);
}

/** decrements refcount of element f and deletes f if refcount is zero */
static inline void
fake_put(struct fake *f) {

	if( !f ) return;
	
	if (atomic_dec_and_test(&f->refcount)) {
		WRITE_LOCK(&fake_lock);
		if(!fake_del(f)) {
			printk(KERN_ERR "arp_arpfake: could not find interface-fake info in list\n");
		}
		WRITE_UNLOCK(&fake_lock);
	}
}


/** tries to find fake-element with actelem->addr == srcip
 *  returns 0 if there were no errors. probably only called from 'fake_find_get'
 *  returns NULL if we didn't find any matching element
 *  precondition: fake-list must be read locked  
 */
static struct fake *
fake_find(u_int32_t srcip)
{
	struct fake *actelem;	

	MUST_BE_READ_LOCKED(&fake_lock);

	if(!fake_head)
		return NULL;
	
	actelem = fake_head;
	
	do
	{
		if( actelem->addr == srcip )
			return actelem;
		actelem = actelem->next;
	}
	while(actelem != NULL);

	return NULL;
}

/** tries to find fake-element with actelem->addr == srcip and increments the
 *  refcount
 *  returns 0 if there were no errors. probably only called from 'fake_find_get'
 *  returns NULL if we didn't find any matching element
 *  postcondition: if returned element isn't used any more fake_put must be
 *                 called  
 */
static struct fake *
fake_find_get(u_int32_t srcip)
{
	struct fake *f;

	READ_LOCK(&fake_lock);
	f = fake_find(srcip);
	if(!f) {
		READ_UNLOCK(&fake_lock);
		return NULL;
	}
	fake_get(f);
	READ_UNLOCK(&fake_lock);
	
	return f;
}


/** probably only called from init_or_cleanup(0) 
 *  FIXME: busywaiting is performed. try to sleep
 */
static void 
fake_free(void)
{
	WRITE_LOCK(&fake_lock);

	while(fake_head) {
		fake_put(fake_head);
	}

	WRITE_UNLOCK(&fake_lock);
}

/*********************************************************************** 
 * DEVICE NOTIFICATION
 ***********************************************************************/

/** helper routine for the notifier functions. in_dev must be rcu_read_locked!!! */
static void
ifa_add_remove(unsigned long event, struct net_device *dev, struct in_ifaddr *ifa)
{
	struct fake *fake_elem;
	u_int8_t mmacaddr[6];
	
	if (event == NETDEV_UP) {
	
		/* FIXME: locking for dev? */
		memcpy(mmacaddr , dev->dev_addr, 6 * sizeof(u_int8_t));
		
		/* by default we take the mac addr from the device and
		 * simply set the multicast bit */
		mmacaddr[0] = mmacaddr[0]|0x01;
		DEBUGP(KERN_DEBUG"arp_arpfake: mmacaddr %02X:%02X:%02X:%02X:%02X:%02X\n", 
		        mmacaddr[0], mmacaddr[1], mmacaddr[2],
		        mmacaddr[3], mmacaddr[4], mmacaddr[5]);
		DEBUGP(KERN_DEBUG"arp_arpfake: addr %u.%u.%u.%u\n", NIPQUAD(ifa->ifa_local));
		if(!fake_add(mmacaddr, ntohl(ifa->ifa_local), mangle))
			printk(KERN_ERR "arp_arpfake: could not add new interface-fake entry %u.%u.%u.%u\n",
			       NIPQUAD(ifa->ifa_local));

	}
	if (event == NETDEV_DOWN) {

		fake_elem = fake_find_get( ntohl(ifa->ifa_local) );
		
		DEBUGP(KERN_DEBUG"arp_arpfake: addr %u.%u.%u.%u\n", NIPQUAD(ifa->ifa_local));
		
		fake_put( fake_elem );
		fake_put( fake_elem );

	}
}
 
static int
arpfake_device_event(struct notifier_block *this,
			     unsigned long event,
			     void *ptr)
{
	struct net_device *dev = ptr;
	struct in_device *in_dev;
	struct in_ifaddr *ifa;
	
	/* we only care about ethernet devices */
	if( dev->type != ARPHRD_ETHER )
		return NOTIFY_DONE;

	if ( event != NETDEV_UP && event != NETDEV_DOWN ) 
		return NOTIFY_DONE;
	
	if ((in_dev = __in_dev_get(dev)) != NULL) {
		for (ifa = in_dev->ifa_list; ifa; ifa = ifa->ifa_next) { 
			DEBUGP(KERN_DEBUG "arp_arpfake: %s interface %s\n",
			        event==NETDEV_UP ? "adding" : "removing", dev->name);
			ifa_add_remove( event, dev, ifa );
		}
	}
	
	return NOTIFY_DONE;
}

static int
arpfake_inet_event(struct notifier_block *this,
			      unsigned long event,
			      void *ptr)
{
	struct net_device *dev = ((struct in_ifaddr *)ptr)->ifa_dev->dev;
	struct in_ifaddr *ifa = ptr;
	
	/* we only care about ethernet devices */
	if( dev->type != ARPHRD_ETHER )
		return NOTIFY_DONE;

	if ( event != NETDEV_UP && event != NETDEV_DOWN ) 
		return NOTIFY_DONE;
	
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0))
//	rcu_read_lock();
#endif
	DEBUGP(KERN_DEBUG "arp_arpfake: %s interface %s\n",
	        event==NETDEV_UP ? "adding" : "removing", dev->name);
	ifa_add_remove( event, dev, ifa );
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0))
//	rcu_read_unlock();
#endif

	return NOTIFY_DONE;
}


static struct notifier_block arpfake_dev_notifier = {
	.notifier_call	= arpfake_device_event,
};

static struct notifier_block arpfake_inet_notifier = {
	.notifier_call	= arpfake_inet_event,
};

/** PREROUTING netfilter hook itself.
  * A unicast IP and a multicast MAC in one packet leads to a packet drop by 
  * TCP respectively routing code. This hook tries to correct those packets
  * which have a corresponding entry in the fakes-list. It rewrites 
  * (*pskb)->pkt_type to PACKET_HOST.
 */
unsigned int pkttype_correct_in(unsigned int hooknum,
			        struct sk_buff **pskb,
			        const struct net_device *in,
			        const struct net_device *out,
			        int (*okfn)(struct sk_buff *))
{
	struct iphdr *iph = (*pskb)->nh.iph;
	
	if( (*pskb)->pkt_type != PACKET_MULTICAST ) {
//		DEBUGP(KERN_DEBUG "arp_arpfake: no multicast mac; we don't correct this packet\n");
		return NF_ACCEPT;
	}
	
	/* we check if the daddr is a multicast ip */
	if( is_mc_ip(iph->daddr) ) {
//		DEBUGP(KERN_DEBUG "arp_arpfake: multicast mac and mc ip, we don't correct this packet\n");
		return NF_ACCEPT;
	}
	
	/* Ok, it's a multicast packet with unicast dst-ip, we correct it */
//	DEBUGP(KERN_DEBUG "arp_arpfake: multicast mac but unicast ip, we correct this packet\n");
	(*pskb)->pkt_type = PACKET_HOST;
	return NF_ACCEPT;
}

/* ARP_OUT netfilter hook itself. Fakes all outgoing arp requests/responses
 * which have a corresponding entry in the fakes-list. 
 */
static unsigned int
arp_mangle(unsigned int hook,
	   struct sk_buff **pskb,
	   const struct net_device *in,
	   const struct net_device *out,
	   int (*okfn)(struct sk_buff *))
{
	struct arphdr *arp = (*pskb)->nh.arph;
	struct arp_payload *payload;
	struct fake *fake_elem;

	/* we don't care about non-ethernet and non-ipv4 ARP */
	if (arp->ar_hrd != htons(ARPHRD_ETHER)
	    || arp->ar_pro != htons(ETH_P_IP)
	    || arp->ar_pln != 4 || arp->ar_hln != ETH_ALEN) {
		DEBUGP(KERN_DEBUG "arp_arpfake: non ipv4 ARP or non ethernet-arp\n");
		return NF_ACCEPT;
	}

	/* we only want to mangle arp requests/replies */
	if (ntohs(arp->ar_op) != ARPOP_REQUEST &&
		ntohs(arp->ar_op) != ARPOP_REPLY) {
		DEBUGP(KERN_DEBUG "arp_arpfake: no ARP request/reply\n");
		return NF_ACCEPT;
	}

	payload = (void *)(arp+1);


	/* normally the linux kernel always replies to arp queries of 
	 * addresses on different interfacs.  However, in the CLUSTERIP case
	 * this wouldn't work, since we didn't subscribe the mcast group on
	 * other interfaces */
//	if (c->dev != out) {
//		DEBUGP("CLUSTERIP: not mangling arp reply on different "
//		       "interface: cip'%s'-skb'%s'\n", c->dev->name, out->name);
//		return NF_ACCEPT;
//	}

	/* mangle src hardware address */
	fake_elem = fake_find_get(ntohl(payload->src_ip));
	if(!fake_elem) {
		DEBUGP(KERN_ERR "arp_arpfake: no multicast-mac in database; %u.%u.%u.%u\n",
		       NIPQUAD(payload->src_ip));
		return NF_ACCEPT;
	}

	if(fake_elem->active)
		memcpy(payload->src_hw , fake_elem->mmacaddr, 6 * sizeof(u_int8_t));
	fake_put(fake_elem);

#ifdef DEBUG_ARPFAKE
	DEBUGP(KERN_DEBUG "arp_arpfake: mangled arp request/reply: ");
	arp_print(payload);
#endif

	return NF_ACCEPT;
}

static struct nf_hook_ops arp_arpfake_in_ops = {
	.hook		= pkttype_correct_in,
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0))
//	.owner		= THIS_MODULE,
#endif
	.pf		= PF_INET,
	.hooknum	= NF_IP_PRE_ROUTING,
	.priority	= NF_IP_PRI_ARPFAKE
};

static struct nf_hook_ops arp_ops = {
	.hook = arp_mangle,
	.pf = NF_ARP,
	.hooknum = NF_ARP_OUT,
	.priority = -1
};

/*********************************************************************** 
 * PROC DIR HANDLING 
 ***********************************************************************/

#ifdef CONFIG_PROC_FS

static int arpfake_proc_read( char *buf, char **start, off_t offset,
                              int size, int *eof, void *data)
{
	int ret;
	struct fake *f = data;

	ret = fake_print( f, buf, size );
	
	if( ret == -ENOMEM ) {
		printk(KERN_ERR "arp_arpfake: could not write entire proc-string\n");
	}
	if( ret == -EINVAL ) {
		printk(KERN_ERR "arp_arpfake: null pointer assignment while printing fake-info\n");
	}
	

	*eof = 1;
	return ret;
}

static int arpfake_proc_write( struct file *instanz, const char __user *userbuffer,
        unsigned long count, void *data )
{
	char *kernel_buffer;
	int not_copied;
	struct fake *f = data;
	u_int8_t mac[6];

	kernel_buffer = kmalloc( count, GFP_KERNEL );
	if( !kernel_buffer ) {
		return -ENOMEM;
	}
	
	not_copied = copy_from_user( kernel_buffer, userbuffer, count );

	if( count == 2 ) {
		WRITE_LOCK(&fake_lock);
		if( *kernel_buffer == '0' ) {
			f->active = 0;
		}
		else if( *kernel_buffer == '1' ) {
			f->active = 1;
		}
		WRITE_UNLOCK(&fake_lock);
		goto finish;
	}
	
	if( count != 18 ) {
		DEBUGP(KERN_DEBUG "arp_arpfake: no MAC entered\n");
		goto finish;
	}
	else {
		*(kernel_buffer+17) = '\0';
	}
	
	if( parse_mac( kernel_buffer, mac ) == -EINVAL ) {
		DEBUGP(KERN_DEBUG "arp_arpfake: invalid MAC entered\n");
		goto finish;
	}
	
	if( !(mac[0]&1) ) {
		DEBUGP(KERN_DEBUG "arp_arpfake: no multicast MAC entered\n");
		goto finish;
	}
	
	WRITE_LOCK(&fake_lock);

	/* we disable the old multicast mac address of the interface... */
	disable_mc( f->addr, f->mmacaddr );
	
	memcpy( f->mmacaddr, mac, 6 * sizeof(u_int8_t) );
	
	/* ...and enable the new multicast mac address of the interface */
	enable_mc( f->addr, f->mmacaddr );
	
	WRITE_UNLOCK(&fake_lock);
	
finish:
	kfree( kernel_buffer );
	return count-not_copied;
}

#endif /* CONFIG_PROC_FS */

static int init_or_cleanup(int fini)
{
	int ret = 0, err;

	if (fini)
		goto cleanup;

	/* NF_ARP_OUT hook registration */
	if (nf_register_hook(&arp_ops) < 0) {
		ret = -EINVAL;
		printk(KERN_ERR "arp_arpfake: can't register ARP_OUT selection hook.\n");
		goto cleanup_nothing;
	}
	
	/* NF_IP_PRE_ROUTING hook registration */
	if(nf_register_hook(&arp_arpfake_in_ops)) {
		ret = -EINVAL;
		printk(KERN_ERR "arp_arpfake: can't register PRE_ROUTING selection hook.\n");
		goto cleanup_arp_hook;
	}

	
	/* Register for device down reports */
	err = register_netdevice_notifier(&arpfake_dev_notifier);

#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0))
	struct net_device *dev;
	read_lock(&dev_base_lock);
	if (!err) {
	for (dev = dev_base; dev; dev = dev->next) {
		(&arpfake_dev_notifier)->notifier_call(&arpfake_dev_notifier, NETDEV_REGISTER, dev);
			if (dev->flags & IFF_UP) 
			(&arpfake_dev_notifier)->notifier_call(&arpfake_dev_notifier, NETDEV_UP, dev);
		}
	}
	read_unlock(&dev_base_lock);
#endif
	register_inetaddr_notifier(&arpfake_inet_notifier);
	/* notifier functions are registered -> fakes will be created */
	
#ifdef CONFIG_PROC_FS
	arpfake_procdir = proc_mkdir("arp_arpfake", NULL);
	if (!arpfake_procdir) {
		printk(KERN_ERR "arp_arpfake: Unable to create proc dir entry 'arp_arpfake'\n");
		ret = -ENOMEM;
		goto cleanup_fakes;
	}
#endif /* CONFIG_PROC_FS */
	
	printk(KERN_DEBUG "ARPFake Version %s loaded successfully\n",
		ARPFAKE_VERSION);

	return 0;

cleanup:
	printk(KERN_DEBUG "ARPFake Version %s unloading\n",
		ARPFAKE_VERSION);
#ifdef CONFIG_PROC_FS
	remove_proc_entry(arpfake_procdir->name, arpfake_procdir->parent);
cleanup_fakes:
#endif /* CONFIG_PROC_FS */
	fake_free();
	unregister_inetaddr_notifier(&arpfake_inet_notifier);
	unregister_netdevice_notifier(&arpfake_dev_notifier);
	nf_unregister_hook(&arp_arpfake_in_ops);
cleanup_arp_hook:
	nf_unregister_hook(&arp_ops);
cleanup_nothing:
	return ret;
}

static int __init init(void)
{
	return init_or_cleanup(0);
}

static void __exit fini(void)
{
	init_or_cleanup(1);
}

module_init(init);
module_exit(fini);
